# Copyright 2024 the LlamaFactory team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import sys
from concurrent.futures import ThreadPoolExecutor

from .constants import RUNNING_LOG


class LoggerHandler(logging.Handler):
    r"""
    Logger handler used in Web UI.
    """

    def __init__(self, output_dir: str) -> None:
        super().__init__()
        formatter = logging.Formatter(
            fmt="%(asctime)s - %(levelname)s - %(name)s - %(message)s", datefmt="%m/%d/%Y %H:%M:%S"
        )
        self.setLevel(logging.INFO)
        self.setFormatter(formatter)

        os.makedirs(output_dir, exist_ok=True)
        self.running_log = os.path.join(output_dir, RUNNING_LOG)
        if os.path.exists(self.running_log):
            os.remove(self.running_log)

        self.thread_pool = ThreadPoolExecutor(max_workers=1)

    def _write_log(self, log_entry: str) -> None:
        with open(self.running_log, "a", encoding="utf-8") as f:
            f.write(log_entry + "\n\n")

    def emit(self, record) -> None:
        if record.name == "httpx":
            return

        log_entry = self.format(record)
        self.thread_pool.submit(self._write_log, log_entry)

    def close(self) -> None:
        self.thread_pool.shutdown(wait=True)
        return super().close()


# def get_logger(name: str) -> logging.Logger:
#     r"""
#     Gets a standard logger with a stream hander to stdout.
#     """
#     formatter = logging.Formatter(
#         fmt="%(asctime)s - %(levelname)s - %(name)s - %(message)s", datefmt="%m/%d/%Y %H:%M:%S"
#     )
#     handler = logging.StreamHandler(sys.stdout)
#     handler.setFormatter(formatter)

#     logger = logging.getLogger(name)
#     logger.setLevel(logging.INFO)
#     logger.addHandler(handler)

#     return logger
from datetime import datetime
def get_logger(name: str) -> logging.Logger:  
    """  
    Gets a standard logger with a stream handler to stdout and file handler with timestamped filename.  
    """  
    formatter = logging.Formatter(  
        fmt="%(asctime)s - %(levelname)s - %(name)s - %(message)s", datefmt="%m/%d/%Y %H:%M:%S"  
    )  
      
    # Stream handler for stdout  
    stream_handler = logging.StreamHandler(sys.stdout)  
    stream_handler.setFormatter(formatter)  
      
    logger = logging.getLogger(name)  
    logger.setLevel(logging.INFO)  
      
    # Add the stream handler to the logger  
    logger.addHandler(stream_handler)  
      
    # Create a log file with a timestamp  
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")  
    log_file = f'./logs/debug_log.txt'  
      
    # If log_file is provided, add a file handler  
    if log_file:  
        os.makedirs(os.path.dirname(log_file), exist_ok=True)  
        file_handler = logging.FileHandler(log_file, encoding='utf-8')  
        file_handler.setFormatter(formatter)  
        logger.addHandler(file_handler)  
      
    return logger  


def reset_logging() -> None:
    r"""
    Removes basic config of root logger. (unused in script)
    """
    root = logging.getLogger()
    list(map(root.removeHandler, root.handlers))
    list(map(root.removeFilter, root.filters))
